from typing import Union, List, Tuple
from tokens import TokenList
from datetime import datetime, date

ChronoResult = Union[date, datetime, Tuple[datetime, datetime]]
ChronoResultList = List[ChronoResult]


class ChronoInterpreter:

    def __init__(self, debug: bool = False):
        self.debug = debug

    def process(self, tokens: TokenList) -> ChronoResultList:
        for t in tokens:
            if self.debug:
                print(t)
            for dt in t.get_chrono_results():
                yield dt



from processors import (
    Processor,
    RegexProcessor,
    IntegerProcessor,
    RemoveIgnoredTokensProcessor,
    ExceptionParsingError,
)
from tokens import (
    RawStringToken,
    TokenList,
    DateToken,
    PartialDateToken,
    TimeToken,
    PartialTimeToken,
    TimeIntervalToken,
    DateTimeToken,
    SlotToken,
    TokenLanguagePack,
)
from utils import remove_accents
import re
from datetime import date, time, datetime, timedelta


class IntegerFrProcessor(IntegerProcessor):
    suffixes = ["er", "e", "eme"]


class DateFrProcessor(Processor):

    def guess_internal(text, list):
        t = remove_accents(text).lower()
        for i, m in enumerate(list):
            if t.startswith(m):
                return i + 1
        return None

    def guess_weekday(text):
        days = [
            "lu",
            "ma",
            "me",
            "je",
            "ve",
            "sa",
            "di",
        ]
        return DateFrProcessor.guess_internal(text, days)

    def guess_month(text):
        mths = [
            "jan",
            "fe",
            "mar",
            "av",
            "mai",
            "juin",
            "juill",
            "ao",
            "sep",
            "oct",
            "nov",
            "dec",
        ]
        return DateFrProcessor.guess_internal(text, mths)

    def parse_french_date(text):
        prefix = None
        weekday = None
        day = None
        month = None
        year = None
        suffix = None
        # format NomJour Numero Mois Année
        m = re.search(
            (
                "(.*?)([a-zA-ZéÉûÛ:.]+)[  ]*([0-9]+)[er]*[  ]*"
                "([a-zA-ZéÉûÛ:.]+)[  ]*([0-9]+)(.*?)"
            ),
            text,
        )
        if m:
            prefix = m.group(1).strip()
            weekday = DateFrProcessor.guess_weekday(m.group(2))
            day = m.group(3)
            month = DateFrProcessor.guess_month(m.group(4))
            year = m.group(5)
            suffix = m.group(6).strip()
        else:
            # format Numero Mois Annee
            m = re.search(
                "(.*?)([0-9]+)[er]*[  ]*([a-zA-ZéÉûÛ:.]+)[  ]*([0-9]+)(.*?)",
                text,
            )
            if m:
                prefix = m.group(1).strip()
                day = m.group(2)
                month = DateFrProcessor.guess_month(m.group(3))
                year = m.group(4)
                suffix = m.group(5).strip()
            else:
                # format Numero Mois Annee
                m = re.search("(.*?)([0-9]+)/([0-9]+)/([0-9]+)(.*?)", text)
                if m:
                    prefix = m.group(1).strip()
                    day = m.group(2)
                    month = int(m.group(3))
                    year = m.group(4)
                    suffix = m.group(5).strip()
                else:
                    # format Numero Mois
                    m = re.search(
                        "(.*?)([0-9]+)[er]*[  ]*([a-zA-ZéÉûÛ:.]+)(.*?)", text
                    )
                    if m:
                        prefix = m.group(1).strip()
                        day = m.group(2)
                        month = DateFrProcessor.guess_month(m.group(3))
                        year = None
                        prefix = m.group(4).strip()
                    else:
                        # format NomJour Numero
                        m = re.search(
                            ("(.*?)([a-zA-ZéÉûÛ:.]+)[  ]*([0-9]+)[er]*(.*?)"),
                            text,
                        )
                        if m:
                            prefix = m.group(1).strip()
                            weekday = DateFrProcessor.guess_weekday(m.group(2))
                            day = None if weekday is None else m.group(3)
                            month = None
                            year = None
                            suffix = m.group(4).strip()

        if day is not None:
            try:
                day = int(day)
            except Exception:
                day = None
            if day >= 32:
                day = None

        if year is not None:
            try:
                year = int(year)
            except Exception:
                year = None

        return year, month, day, weekday, prefix, suffix

    @classmethod
    def process(
        cls, input: TokenList, pack: TokenLanguagePack, context: date
    ) -> TokenList:
        for token in input:
            if isinstance(token, RawStringToken):
                year, month, day, weekday, prefix, suffix = (
                    DateFrProcessor.parse_french_date(token.text)
                )
                if year is not None and month is not None and day is not None:
                    if prefix is not None and len(prefix) > 0:
                        yield RawStringToken(prefix)
                    yield DateToken(date(year, month, day), token.text)
                    if suffix is not None and len(suffix) > 0:
                        yield RawStringToken(suffix)
                elif (
                    month is not None or weekday is not None
                ) and day is not None:
                    if prefix is not None and len(prefix) > 0:
                        yield RawStringToken(prefix)
                    yield PartialDateToken(
                        year=year, month=month, day=day, raw=token.text
                    )
                    if suffix is not None and len(suffix) > 0:
                        yield RawStringToken(suffix)
                else:
                    yield token
            else:
                yield token


class TimeFrProcessor(Processor):

    def parse_french_time(text):

        prefix = None
        h = None
        m = None
        s = None
        suffix = None

        # format heures minutes secondes
        mx = re.search(
            "(.*?)([0-9]+)[ a-zA-Z:.]+([0-9]+)" "[ a-zA-Z:.]+([0-9]+)(.*?)",
            text,
        )
        if mx:
            prefix = mx.group(1)
            h = mx.group(2)
            m = mx.group(3)
            s = mx.group(4)
            suffix = mx.group(5)
        else:
            # format heures minutes
            mx = re.search("(.*?)([0-9]+)[ hH:.]+([0-9]+)(.*?)", text)
            if mx:
                prefix = mx.group(1)
                h = mx.group(2)
                m = mx.group(3)
                s = "0"
                suffix = mx.group(4)
            else:
                # format heures
                mx = re.search("(.*?)([0-9]+)[ ]*[Hh:.](.*?)", text)
                if mx:
                    prefix = mx.group(1)
                    h = mx.group(2)
                    m = "0"
                    s = "0"
                    suffix = mx.group(3)

        if h is not None and m is not None and s is not None:
            try:
                h = int(h)
                m = int(m)
                s = int(s)
            except Exception:
                h = None
                m = None
                s = None
        if h is not None and h >= 24:
            h = None
        if m is not None and m >= 60:
            m = None
        if s is not None and s >= 60:
            s = None
        return h, m, s, prefix, suffix

    @classmethod
    def process(
        cls, input: TokenList, pack: TokenLanguagePack, context: date
    ) -> TokenList:
        for token in input:
            if isinstance(token, RawStringToken):
                h, m, s, prefix, suffix = TimeFrProcessor.parse_french_time(
                    token.text
                )
                if h is not None and m is not None and s is not None:
                    if prefix is not None and len(prefix) > 0:
                        yield RawStringToken(prefix)
                    yield TimeToken(time(h, m, s), token.text)
                    if suffix is not None and len(suffix) > 0:
                        yield RawStringToken(suffix)
                else:
                    yield token
            else:
                yield token


class TimeIntervalProcessor(RegexProcessor):
    regex = ":(1|t|T)(-|à)T"

    def convert_match(cls, tokens: TokenList, context: date) -> TokenList:
        times = [
            t for t in tokens if isinstance(t, (TimeToken, PartialTimeToken))
        ]
        if len(times) == 2:
            t1 = times[0]
            t2 = times[1]
            if isinstance(t1, PartialTimeToken):
                t1 = t1.consolidate_time(context=t2.time)
            yield TimeIntervalToken(t1.time, t2.time)
        else:
            raise ExceptionParsingError(tokens)


class DateIntervalProcessor(RegexProcessor):
    regex = "fDoD"

    def convert_match(cls, tokens: TokenList, context: date) -> TokenList:
        dates = [t for t in tokens if isinstance(t, DateToken)]
        if len(dates) == 2:
            d1 = dates[0].date
            d2 = dates[1].date
            while d1 <= d2:
                yield DateToken(d1, "")
                d1 = d1 + timedelta(days=1)
        else:
            raise ExceptionParsingError(tokens)


class DateConsolidationProcessor(RegexProcessor):
    regex = "((1|d)[^D]*)D"

    def convert_match(cls, tokens: TokenList, context: date) -> TokenList:
        dt = [t for t in tokens if isinstance(t, DateToken)]
        if len(dt) == 1:
            for d in tokens:
                if isinstance(d, PartialDateToken):
                    yield d.consolidate_date(context=dt[0].date)
                else:
                    yield d
        else:
            raise ExceptionParsingError(tokens)


class ReverseDateConsolidationProcessor(DateConsolidationProcessor):
    regex = "D((1|d)[^D]*)"


class DateConsolidationByContextProcessor(RegexProcessor):
    regex = ".*"

    def convert_match(cls, tokens: TokenList, context: date) -> TokenList:
        tlist = list(tokens)
        current = context

        # update starting from the last one
        for i in reversed(range(len(tlist))):
            if isinstance(tlist[i], PartialDateToken):
                tlist[i] = tlist[i].consolidate_date(context=current)
                current = tlist[i].date

        for obj in tlist:
            yield obj


class DateListTimeListProcessor(RegexProcessor):
    regex = "D+(à|§)?(T|i)+"

    def convert_match(cls, tokens: TokenList, context: date) -> TokenList:
        dates = [t for t in tokens if isinstance(t, DateToken)]
        times = [
            t for t in tokens if isinstance(t, (TimeToken, TimeIntervalToken))
        ]

        if len(dates) != 0 and len(times) != 0:
            for d in dates:
                for t in times:
                    if isinstance(t, TimeToken):
                        yield DateTimeToken(datetime.combine(d.date, t.time))
                    else:
                        if isinstance(t, TimeIntervalToken):
                            if t.start > t.end:
                                yield SlotToken(
                                    datetime.combine(d.date, t.start),
                                    datetime.combine(
                                        d.date + timedelta(days=1), t.end
                                    ),
                                )
                            else:
                                yield SlotToken(
                                    datetime.combine(d.date, t.start),
                                    datetime.combine(d.date, t.end),
                                )
                        else:
                            raise ExceptionParsingError(t)
        else:
            raise ExceptionParsingError(tokens)


class SlotProcessor(RegexProcessor):

    regex = "(f)?ror"

    def convert_match(cls, tokens: TokenList, context: date) -> TokenList:
        datetimes = [t for t in tokens if isinstance(t, DateTimeToken)]

        if len(datetimes) == 2:
            yield SlotToken(datetimes[0].datetime, datetimes[1].datetime)
        else:
            raise ExceptionParsingError(tokens)


french_processors = [
    # cleaning
    RemoveIgnoredTokensProcessor,
    # build date and time
    DateFrProcessor,
    TimeFrProcessor,
    IntegerFrProcessor,
    # detect time intervals
    TimeIntervalProcessor,
    # consolidate dates
    DateConsolidationProcessor,
    ReverseDateConsolidationProcessor,
    DateConsolidationByContextProcessor,
    # detect date intervals
    DateIntervalProcessor,
    # multiple datetimes
    DateListTimeListProcessor,
    # detect slots
    SlotProcessor,
]



from tokens import (
    TokenLanguagePack,
    AdditiveConjunctionToken,
    NotSupportedToken,
    IgnoredToken,
    IntervalConjunctionToken,
    PrecisionConjunctionToken,
)


class FrenchTokens(TokenLanguagePack):
    class EtToken(AdditiveConjunctionToken):
        substr = "et"

    class VirguleToken(AdditiveConjunctionToken):
        substr = ","

    class PuisToken(AdditiveConjunctionToken):
        substr = "puis"

    class OuToken(NotSupportedToken):
        substr = "ou"

    class SaufToken(NotSupportedToken):
        substr = "sauf"

    class LeToken(IgnoredToken):
        substr = "le"

    class LesToken(IgnoredToken):
        substr = "les"

    class DuToken(IntervalConjunctionToken):
        substr = "du"
        class_char = "f"

    class DeToken(IntervalConjunctionToken):
        substr = "de"
        class_char = ":"

    class AuToken(IntervalConjunctionToken):
        substr = "au"
        class_char = "o"

    class JusquAuToken(IntervalConjunctionToken):
        substr = "jusqu'au"
        class_char = "o"

    class AToken(PrecisionConjunctionToken):
        substr = "à"
        class_char = "à"

    class DashToken(PrecisionConjunctionToken):
        substr = "-"
        class_char = "-"

    conjunctions = [
        PuisToken,
        DeToken,
        DuToken,
        AuToken,
        JusquAuToken,
        AToken,
        DashToken,
    ]

    ignored_conjunctions = [LeToken, LesToken, EtToken, VirguleToken]

    not_supported = [OuToken, SaufToken]



from abc import ABC
from typing import List, Type
from datetime import date, time, datetime, timedelta


class Token(ABC):

    class_char: str

    def __init_subclass__(cls):
        super().__init_subclass__()
        if not getattr(cls, "abstract", False) and (
            not hasattr(cls, "class_char")
            or not isinstance(cls.class_char, str)
        ):
            raise TypeError(
                f"{cls.__name__} must define a class-level "
                "'class_char' of type str."
            )

    def get_chrono_results(self):
        return []


TokenList = List[Token]


class DateToken(Token):

    class_char = "D"

    def __init__(self, date: date, raw: str):
        self._date = date
        self.raw = raw

    @property
    def date(self) -> date:
        return self._date

    def __str__(self):
        return "<" + self.__class__.__name__ + ":" + str(self._date) + ">"

    def get_chrono_results(self):
        return [self._date]


class TimeToken(Token):

    class_char = "T"

    def __init__(self, time: time, raw: str):
        self._time = time
        self.raw = raw

    @property
    def time(self) -> time:
        return self._time

    def __str__(self):
        return "<" + self.__class__.__name__ + ":" + str(self._time) + ">"

    def get_chrono_results(self):
        return [self._time]


class TimeIntervalToken(Token):

    class_char = "i"

    def __init__(self, time1, time2: time, src: TokenList = None):
        self._time1 = time1
        self._time2 = time2
        self.src = src

    @property
    def start(self) -> time:
        return self._time1

    @property
    def end(self) -> time:
        return self._time2

    def __str__(self):
        return (
            "<"
            + self.__class__.__name__
            + ":"
            + str(self.start)
            + " - "
            + str(self.end)
            + ">"
        )

    def get_chrono_results(self):
        return [(self.start, self.end)]


class DateTimeToken(Token):

    class_char = "r"

    def __init__(self, datetime: datetime, src: TokenList = None):
        self._datetime = datetime
        self.src = src

    @property
    def datetime(self) -> datetime:
        return self._datetime

    def __str__(self):
        return "<" + self.__class__.__name__ + ":" + str(self.datetime) + ">"

    def get_chrono_results(self):
        return [self.datetime]


class SlotToken(Token):

    class_char = "s"

    def __init__(self, datetime1, datetime2: datetime, src: TokenList = None):
        self._datetime1 = datetime1
        self._datetime2 = datetime2
        self.src = src

    @property
    def start(self) -> datetime:
        return self._datetime1

    @property
    def end(self) -> datetime:
        return self._datetime2

    def __str__(self):
        return (
            "<"
            + self.__class__.__name__
            + ":"
            + str(self.start)
            + "-"
            + str(self.end)
            + ">"
        )

    def get_chrono_results(self):
        return [(self.start, self.end)]


class RawStringToken(Token):

    class_char = "§"

    def __init__(self, text: str):
        self.text = text

    def __str__(self):
        return "<" + self.__class__.__name__ + ":" + self.text + ">"


class PartialTimeToken(RawStringToken):

    class_char = "t"

    def __init__(
        self,
        raw: str,
        hour: int = None,
        minute: int = None,
        second: int = None,
    ):
        super().__init__(raw)
        self.hour = hour
        self.minute = minute
        self.second = second

    def __str__(self):
        return (
            "<"
            + self.__class__.__name__
            + ":"
            + str(self.hour)
            + ":"
            + str(self.minute)
            + ":"
            + str(self.second)
            + " ("
            + str(self.text)
            + ")>"
        )

    def consolidate_time(self, context: time) -> TimeToken:
        t = time(
            context.hour if self.hour is None else self.hour,
            context.minute if self.minute is None else self.minute,
            context.second if self.second is None else self.second,
        )
        return TimeToken(raw=self.text, time=t)


class PartialDateToken(RawStringToken):

    class_char = "d"

    def __init__(
        self,
        raw: str,
        year: int = None,
        month: int = None,
        day: int = None,
        weekday: str = None,
    ):
        super().__init__(raw)
        self.year = year
        self.month = month
        self.day = day
        self.weekday = weekday

    def __str__(self):
        return (
            "<"
            + self.__class__.__name__
            + ":"
            + str(self.year)
            + ", "
            + str(self.month)
            + ", "
            + str(self.day)
            + " ("
            + str(self.text)
            + ")>"
        )

    def consolidate_date(self, context: date) -> DateToken:
        d = date(
            context.year if self.year is None else self.year,
            context.month if self.month is None else self.month,
            context.day if self.day is None else self.day,
        )
        # adjust year by proximity
        if self.year is None:
            dates = [date(d.year + x, d.month, d.day) for x in [-1, 0, 1]]
            dates = [(abs((d - context).days), d) for d in dates]
            d = min(dates, key=lambda x: x[0])[1]

        return DateToken(raw=self.text, date=d)


class IntToken(PartialDateToken, PartialTimeToken):

    class_char = "1"

    def __init__(self, raw: str, value: int):
        PartialDateToken.__init__(self, raw=raw, day=value)
        PartialTimeToken.__init__(self, raw=raw, hour=value)

    def __str__(self):
        return (
            "<"
            + self.__class__.__name__
            + ":"
            + str(self.day)
            + " ("
            + str(self.text)
            + ")>"
        )


class FixedStringToken(Token):
    abstract = True

    substr: str

    def __init_subclass__(cls):
        super().__init_subclass__()
        if not getattr(cls, "abstract", False) and (
            not hasattr(cls, "substr") or not isinstance(cls.substr, str)
        ):
            raise TypeError(
                f"{cls.__name__} must define a class-level "
                "'substr' of type str."
            )

    def __str__(self):
        return "<" + self.__class__.__name__ + ":" + self.substr + ">"


class ConjunctionToken(FixedStringToken):
    abstract = True


class AdditiveConjunctionToken(ConjunctionToken):
    class_char = "&"
    abstract = True


class NotSupportedToken(FixedStringToken):
    class_char = "x"
    abstract = True


class IgnoredToken(FixedStringToken):
    class_char = " "
    abstract = True


class IntervalConjunctionToken(ConjunctionToken):
    class_char = "~"
    abstract = True


class PrecisionConjunctionToken(ConjunctionToken):
    class_char = "+"
    abstract = True


class ExceptionNotSupportedToken(Exception):
    def __init__(self, t: NotSupportedToken):
        self._t = t

    def __str__(self):
        return "ExceptionNotSupportedToken: " + str(self._t)


class TokenLanguagePack(object):

    not_supported: List[Type[NotSupportedToken]]

    ignored_conjunctions: List[Type[IgnoredToken]]

    conjunctions: List[Type[IntervalConjunctionToken]]

    def __init__(self):
        super().__init__()

        self._token_from_str = {}
        for t in self.tokens:
            self._token_from_str[t.substr] = t

    def token_from_str(self, txt: str) -> Token:
        if txt in self._token_from_str:
            t = self._token_from_str[txt]()
            if isinstance(t, NotSupportedToken):
                raise ExceptionNotSupportedToken(t)
            return t
        else:
            return RawStringToken(txt)

    @property
    def tokens(self):
        return (
            self.conjunctions + self.not_supported + self.ignored_conjunctions
        )

    @property
    def token_strings(self):
        return [x.substr for x in self.tokens]


def get_token_pack(lang: str = "fr") -> TokenLanguagePack:
    if lang == "fr":
        from tokens_fr import FrenchTokens

        return FrenchTokens()
    else:
        raise ValueError("Unsupported language")



import os
from setuptools import setup, find_packages

setup(
    name='chronostring',
    version=open('VERSION').read(),
    description='Extracts and structures dates and times from natural French text.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Jean-Marie Favreau',
    url='https://forge.chapril.org/jmtrivial/chronostring',
    author_email='jeanmarie.favreau@free.fr',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Affero General Public License v3',
    ],
    python_requires='>=3.7',
)



from abc import ABC, abstractmethod
import re
from typing import List
from tokens import (
    TokenList,
    TokenLanguagePack,
    RawStringToken,
    IntToken,
)
from datetime import date


class ExceptionParsingError(Exception):
    def __init__(self, t: TokenList):
        self._t = t

    def __str__(self):
        return "ExceptionParsingError: " + str(self._t)


class Processor(ABC):

    def __init__(self, debug: bool = False):
        self.debug = debug

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, d):
        self._debug = d

    @classmethod
    @abstractmethod
    def process(
        cls, input: TokenList, pack: TokenLanguagePack, context: date
    ) -> TokenList:
        pass


class RemoveIgnoredTokensProcessor(Processor):

    @classmethod
    def process(
        cls, input: TokenList, pack: TokenLanguagePack, context: date
    ) -> TokenList:
        for token in input:
            if not isinstance(token, tuple(pack.ignored_conjunctions)):
                yield token


class IntegerProcessor(Processor):
    suffixes = []

    @classmethod
    def is_integer(cls, txt):
        if txt.isdigit():
            return True
        for s in cls.suffixes:
            if txt.endswith(s) and txt[: -len(s)].isdigit():
                return True
        return False

    @classmethod
    def process(
        cls, input: TokenList, pack: TokenLanguagePack, context: date
    ) -> TokenList:
        for token in input:
            if type(token) == RawStringToken and cls.is_integer(token.text):
                v = int(re.sub(r"[^\d\.]", "", token.text))
                if v >= 0:
                    yield IntToken(raw=token.text, value=v)
                else:
                    yield token
            else:
                yield token


class RegexProcessor(Processor, ABC):
    regex: str

    def __init_subclass__(cls):
        super().__init_subclass__()
        if not getattr(cls, "abstract", False) and (
            not hasattr(cls, "regex") or not isinstance(cls.regex, str)
        ):
            raise TypeError(
                f"{cls.__name__} must define a class-level "
                "'regex' of type str."
            )

    def to_charlist(tokens):
        return "".join([t.class_char for t in tokens])

    @classmethod
    @abstractmethod
    def convert_match(cls, tokens: TokenList, context: date) -> TokenList:
        pass

    @classmethod
    def intermediate_tokens(
        cls, tokens: TokenList, start: int, end: int
    ) -> TokenList:
        return tokens[start:end]

    @classmethod
    def process(
        cls, input: TokenList, pack: TokenLanguagePack, context: date
    ) -> TokenList:
        tlist = list(input)
        txt = RegexProcessor.to_charlist(tlist)

        pattern = re.compile(cls.regex)

        last_end = 0
        for match in pattern.finditer(txt):
            if match.start() > last_end:
                yield from cls.intermediate_tokens(
                    tlist, last_end, match.start()
                )
            yield from cls.convert_match(
                cls,
                cls.intermediate_tokens(tlist, match.start(), match.end()),
                context,
            )
            last_end = match.end()
        if last_end < len(txt):
            yield from cls.intermediate_tokens(tlist, last_end, None)


def get_processors(lang: str = "fr") -> List[Processor]:
    if lang == "fr":
        from processors_fr import french_processors

        return french_processors
    else:
        raise ValueError("Unsupported language")
